package filesystem

import java.nio.file.FileAlreadyExistsException
import java.nio.file.Files
import java.nio.file.Path

class FSCreator {
    /**
     * Create entry, leaving existing folders' contents, but overwriting existing files.
     * @throws CyclicFolderException Cyclic folders cannot be created.
     */
    @Throws(CyclicFolderException::class)
    fun create(
        entryToCreate: FSEntry,
        destination: String,
    ) {
        // No point in running anything if we know the input is invalid.
        if (entryToCreate is FSFolder && entryToCreate.isCyclic()) {
            throw CyclicFolderException()
        }

        val queue =
            ArrayDeque(
                listOf(
                    entryToCreate to Path.of(destination),
                ),
            )

        while (queue.isNotEmpty()) {
            val (entry, dest) = queue.removeFirst()
            val path = dest.resolve(entry.name)
            try {
                when (entry) {
                    is FSFile -> Files.createFile(path)
                    is FSFolder -> Files.createDirectory(path)
                }
            } catch (_: FileAlreadyExistsException) {
            } // Allow files/folders to already exist.
            when (entry) {
                is FSFile -> Files.write(path, entry.content.toByteArray())
                is FSFolder -> queue.addAll(entry.entries.map { it to path })
            }
        }
    }
}

class CyclicFolderException : Exception("Cyclic FSFolders are not supported")